"""
bp_broker service broker bundle.

Service registration, listing, and object querying
"""


import re
import time
import json

import bpbroker


#####################################################

def Register(rh):
	"""Registers new service to service broker.

	Registers a new entry if none exists or returns error if entry already exists.
	Recommend using a name + key to mitigate misplaced overwrites

	:param name: Unique registration name.  Often a name and a unique key.
	:param data: json object containing all data to associated with name
	:returns success: bool success
	:returns message: status message
	:returns data: query result for key 'name'
	"""

	# Validate parameters
	data = []
	try:
		data = json.loads(rh.qs['data'])
	except:
		data = {'_str': rh.qs['data']}
	if 'name' not in rh.qs:  
		rh.status = 400
		rh.status_message = "Missing name parameter"
	elif 'data' not in rh.qs:  
		rh.status = 400
		rh.status_message = "Missing data parameter"
	elif 'last_write_ip' in data:  
		rh.status = 400
		rh.status_message = "Used reserved data name last_write_ip"
	elif 'last_write_ts' in data:  
		rh.status = 400
		rh.status_message = "Used reserved data name last_write_ts"

	# Set data
	elif data:  
		with bpbroker.config.rlock:
			if rh.qs['name'] in bpbroker.config.data['services']:  
				rh.data = json.dumps({'success': False, 'message': "Entry '%s' already exists" % rh.qs['name'], 'data': {}})
			else:  
				bpbroker.config.data['services'][rh.qs['name']] = \
					dict(data.items() + {'last_write_ip': rh.RequestingHost(), 'last_write_ts': int(time.time())}.items())
				Get(rh)


def Replace(rh):
	"""Replacing existing content (if any).

	:param name: Unique registration name.  Often a name and a unique key.
	:param data: json object containing all data to associated with name
	:returns success: bool success
	:returns message: status message
	:returns data: query result for key 'name'
	"""
	Delete(rh)
	Register(rh)


def Delete(rh):
	"""Remove keyed entry.

	:param name: Unique registration name.  Often a name and a unique key.
	:returns success: bool success
	:returns message: status message
	"""
	# Validate parameters
	error = False
	if 'name' not in rh.qs:
		rh.status = 400
		rh.status_message = "Missing name parameter"

	# Get data
	else:
		with bpbroker.config.rlock:
			if rh.qs['name'] in bpbroker.config.data['services']:  del(bpbroker.config.data['services'][rh.qs['name']])
			rh.data = json.dumps({'success': True, 'message': "Success"})



def Update(rh):
	"""Update existing entry.

	If no entry exists insert it.  If entry already exists for given key then merge data together with
	new data taking precedence.

	:param name: Unique registration name.  Often a name and a unique key.
	:param data: json object containing all data to associated with name
	:returns success: bool success
	:returns message: status message
	:returns data: query result for key 'name'
	"""

	# Validate parameters
	data = []
	try:
		data = json.loads(rh.qs['data'])
	except:
		data = {'_str': rh.qs['data']}
	if 'name' not in rh.qs:  
		rh.status = 400
		rh.status_message = "Missing name parameter"
	elif 'data' not in rh.qs:  
		rh.status = 400
		rh.status_message = "Missing data parameter"
	elif 'last_write_ip' in data:  
		rh.status = 400
		rh.status_message = "Used reserved data name last_write_ip"
	elif 'last_write_ts' in data:  
		rh.status = 400
		rh.status_message = "Used reserved data name last_write_ts"

	# Set data
	elif data:  
		with bpbroker.config.rlock:
			if rh.qs['name'] not in bpbroker.config.data['services']:  bpbroker.config.data['services'][rh.qs['name']] = {}
			bpbroker.config.data['services'][rh.qs['name']] = \
				dict(bpbroker.config.data['services'][rh.qs['name']].items() + data.items() + 
				     {'last_write_ip': rh.RequestingHost(), 'last_write_ts': int(time.time())}.items())
			Get(rh)


def Get(rh):
	"""Return data associated with given key.

	Returns all data associated with key unless specific fields are provided.

	:param name: Unique registration name.  Often a name and a unique key.
	:returns success: bool success
	:returns message: status message
	:returns data: query result for key 'name'
	"""
	# Validate parameters
	error = False
	if 'name' not in rh.qs:  
		rh.status = 400
		rh.status_message = "Missing name parameter"

	# Get data
	else:
		with bpbroker.config.rlock:
			if rh.qs['name'] not in bpbroker.config.data['services']: 
				rh.data = json.dumps({'success': False, 'message': "Entry '%s' not found" % rh.qs['name'], 'data': {}})
			else:  
				rh.data = json.dumps({'success': True, 'message': "Success", 'data': bpbroker.config.data['services'][rh.qs['name']]})


def List(rh):
	"""Alias for Get."""
	Get(rh)





# -*- coding: utf-8 -*-
"""
CenturyLink Cloud Blueprint Broker Toolset.

This broker foundation works to support advanced Blueprint deployment requiring
shared state and multi-server/multi-role communication across multiple server deployments.

CenturyLink Cloud: http://www.CenturyLinkCloud.com
Package Github page: 

"""

from bpclient.shell import Args, ExecCommand
import bpclient.cli as cli
import bpclient.discover as discover
import bpclient.ping as ping
import bpclient.output as output
import bpclient.services as services
import bpclient.execute as execute


import requests
requests.packages.urllib3.disable_warnings()


####### module/object vars #######
#_V1_API_KEY = False




# -*- coding: utf-8 -*-
"""Output formatting helper functions."""

import os
import sys
import json
import time



# TODO add sorting
def Rows(data_arr,keys,opts={}):
	max_key_len = 0
	for key in keys: 
		if len(key)>max_key_len:  max_key_len = len(key)

	table = ''
	i = 0
	for line in data_arr:
		i += 1
		table += "\n  ******************* %s. ********************\n" % (i)
		for key in keys: table += "%s:  %s\n" % (key.rjust(max_key_len+2),line[key])

	return(table)


# TSV w/o headers
def Text(data,keys,opts={}):
	for key in keys: 
		if isinstance(data[key], (basestring, int, long, float)):  row = str(data[key]).replace("	"," ")
		elif isinstance(data[key], (dict,)):  row = json.dumps(data[key])
		else:  
			str_line = []
			for a in data[key]:  str_line.append(str(a))
			row = ", ".join(str_line).replace(",","")

	return(row)


# TODO - Use CSV module?
def Csv(data,keys,opts={'no_header': False}):
	if not opts['no_header']:  csv.append(",".join(keys))

	for key in keys: 
		if isinstance(data[key], (basestring, int, long, float)):  csv = str(data[key]).replace(","," ")
		else:  
			str_data = []
			for a in data[key]:  str_data.append(str(a))
			csv =" ".join(str_data).replace(",","")

	return(csv)


def Json(data,keys,opts={}):
	for key in data.keys():
		if key not in keys:  data.pop(key,None)

	return(data)




#!/usr/bin/env python

import Queue
import signal
import time
import sys

import bpbroker



####################################################

def Shutdown(signum, fram):
	bpbroker.config.Save()
	print "\n"
	sys.exit(0)

####################################################

def Start():

	queue_worker = Queue.Queue()
	queue_health = Queue.Queue()

	api_thread = bpbroker.API.APIThread(queue_worker,queue_health)
	api_thread.start()

	discover_thread = bpbroker.discover.DiscoverThread(queue_worker,queue_health)
	discover_thread.start()


	signal.signal(signal.SIGINT,Shutdown)
	while True:  time.sleep(5)




"""
bp_broker API module.

Listens and responds to ssl connections.  Proxies connection requests to registered event handlers.
"""


import threading
import Queue
#import json
import SocketServer
import time
#import re
#from urlparse import urlparse, parse_qs, parse_qsl
#from StringIO import StringIO

import bpbroker


default_config = {
	'listen_port': 20443,
	'listen_ip': '',	# default bind to all interfaces
}



#####################################################


class TimeoutBaseUDPServer(SocketServer.UDPServer):
    timeout = 1



class DiscoverUDPHandler(SocketServer.BaseRequestHandler):

	def handle(self):
		data = self.request[0].strip()
		with bpbroker.config.rlock:
			socket = self.request[1]
			if data in bpbroker.config.data['services']:
				socket.sendto("Key Exists\n", self.client_address)



class DiscoverThread(threading.Thread):

	def __init__(self,worker_queue,health_queue,config={}):
		threading.Thread.__init__(self)
		self.worker_queue = worker_queue
		self.health_queue = health_queue
		self._stop_event = threading.Event()
		self.config = dict(list(default_config.items()) + list(config.items()))


	def join(self,timeout=None):
		self._stop_event.set()
		threading.Thread.join(self, timeout)


	def run(self):
		self.discover_server = TimeoutBaseUDPServer((self.config['listen_ip'], self.config['listen_port']), DiscoverUDPHandler)

		while not self._stop_event.is_set():
			self.discover_server.handle_request()
			self.HealthCheck()


	def HealthCheck(self):
		self.health_queue.put_nowait({'thread': 'discover', 'ts': int(time.time())})






"""
bp_broker service broker bundle.

Service registration, listing, and object querying
"""


import re
import time
import json
import requests

import bpclient


#####################################################

def Register(access_key,name,data):
	"""Registers new service to service broker.

	Registers a new entry if none exists or returns error if entry already exists.
	Recommend using a name + key to mitigate misplaced overwrites

	CLI:
	> ./bpclient.py  -f json -b 127.0.0.1:20443 service register --name test2 --data '{"a": 3, "b": "foo"}'
	{"message": "Success", "data": {"a": 3, "last_write_ts": 1418760026, "b": "foo", "last_write_ip": "127.0.0.1"}, "success": true}

	REPL:
	>>> bpclient.services.Register(name='test6',data='{"a": 1}')
	{u'data': {u'a': 1,
	           u'last_write_ip': u'127.0.0.1',
	           u'last_write_ts': 1418760363},
	 u'message': u'Success',
	 u'success': True}

	>>> bpclient.services.Register(name='test5',data='my data')
	{u'data': {u'data': u'my data',
	           u'last_write_ip': u'127.0.0.1',
	           u'last_write_ts': 1418760632},
	 u'message': u'Success',
	 u'success': True}

	:param name: Unique registration name.  Often a name and a unique key.
	:param data: json object containing all data to associated with name
	"""

	r = requests.post("https://%s/services/Register/" % bpclient.BPBROKER,params={'access_key': access_key, 'name': name, 'data': data},verify=False)
	return(r.json())


def Replace(access_key,name,data):
	"""Replacing existing content (if any).

	CLI:
	> ./bpclient.py  -f text -b 127.0.0.1:20443 service register --name test1 --data 'original data'
	True	Success	{"last_write_ts": 1418761404, "data": "original data", "last_write_ip": "127.0.0.1"}
 	>./bpclient.py  -f text -b 127.0.0.1:20443 service replace --name test1 --data 'new data'
	True	Success	{"last_write_ts": 1418761411, "data": "new data", "last_write_ip": "127.0.0.1"}

	REPL:
	>>> bpclient.services.Register(name='test2',data='original data')
	{u'message': u'Success', u'data': {u'last_write_ts': 1418761514, u'data': u'original data', u'last_write_ip': u'127.0.0.1'}, u'success': True}

	>>> bpclient.services.Replace(name='test2',data='new data')
	{u'message': u'Success', u'data': {u'last_write_ts': 1418761523, u'data': u'new data', u'last_write_ip': u'127.0.0.1'}, u'success': True}

	:param name: Unique registration name.  Often a name and a unique key.
	:param data: json object containing all data to associated with name
	"""
	r = requests.post("https://%s/services/Replace/" % bpclient.BPBROKER,params={'access_key': access_key, 'name': name, 'data': data},verify=False)
	return(r.json())


def Delete(access_key,name):
	"""Remove keyed entry.  

	Returns scuccess=True whether key exists or not.

	CLI:
	> ./bpclient.py  -f text -b 127.0.0.1:20443 service delete --name test1
	True	Success

	REPL:
	>>> bpclient.services.Delete('test1')
	{u'message': u'Success', u'success': True}

	:param name: Unique registration name.  Often a name and a unique key.
	"""

	r = requests.post("https://%s/services/Delete/" % bpclient.BPBROKER,params={'access_key': access_key, 'name': name},verify=False)
	return(r.json())


def Update(access_key,name,data):
	"""Update existing entry.

	If no entry exists insert it.  If entry already exists for given key then merge data together with
	new data taking precedence.  Perform deep merge of data if it is a json object.

	CLI:
	> ./bpclient.py  -f text -b 127.0.0.1:20443 service update --name test4 --data '{"old_key": "old_data"}'
	True	Success	{"old_key": "old_data", "last_write_ts": 1418762954, "last_write_ip": "127.0.0.1"}
 	>./bpclient.py  -f text -b 127.0.0.1:20443 service update --name test4 --data '{"new_key": "new_data"}'
	True	Success	{"old_key": "old_data", "last_write_ts": 1418762962, "last_write_ip": "127.0.0.1", "new_key": "new_data"}

	REPL:
	>>> bpclient.services.Update(name='test3',data='{"orig_key": "orig_data"}')
	{u'data': {u'last_write_ip': u'127.0.0.1',
	           u'last_write_ts': 1418762718,
	           u'orig_key': u'orig_data'},
	 u'message': u'Success',
	 u'success': True}

	>>> bpclient.services.Update(name='test3',data='{"new_key": "new_data"}')
	{u'data': {u'last_write_ip': u'127.0.0.1',
	           u'last_write_ts': 1418762729,
	           u'new_key': u'new_data',
	           u'orig_key': u'orig_data'},
	 u'message': u'Success',
	 u'success': True}

	:param name: Unique registration name.  Often a name and a unique key.
	:param data: json object containing all data to associated with name
	"""

	r = requests.post("https://%s/services/Update/" % bpclient.BPBROKER,params={'access_key': access_key, 'name': name, 'data': data},verify=False)
	return(r.json())


def Get(access_key,name):
	"""Return data associated with given key.

	Returns all data associated with key unless specific fields are provided.

	CLI:
	> ./bpclient.py  -f text -b 127.0.0.1:20443 service get --name test2
	True	Success	{"last_write_ts": 1418760685, "data": "xxxx", "last_write_ip": "127.0.0.1"}

	REPL:
	>>> bpclient.services.Get('test1')
	{u'data': {u'data': u'xxxx',
	           u'last_write_ip': u'127.0.0.1',
	           u'last_write_ts': 1418760626},
	 u'message': u'Success',
	 u'success': True}

	>>> bpclient.services.Get('nokey')
	{u'message': u'Entry not found', u'data': {}, u'success': False}

	:param name: Unique registration name.  Often a name and a unique key.
	"""

	r = requests.post("https://%s/services/Get/" % bpclient.BPBROKER,params={'access_key': access_key, 'name': name},verify=False)
	return(r.json())


def List(access_key,name):
	"""Alias for Get."""
	Get(access_key,name)


